import {distance} from './helpers';

export class Plant extends Phaser.Group {
    constructor(game) {
        super(game, undefined, 'plant', true);

        this.stem = [new Stem(game)];
        this.stemBitmap = new Phaser.BitmapData(game, 'stem', game.width, game.height - game.height / 3);
        this.leafPositions = [];

        this.roots = [new Root(game), new Root(game), new Root(game)];
        this.rootBitmap = new Phaser.BitmapData(game, 'roots', game.width, game.height / 3);

        this.stemsGrowing = 1;
        this.rootsGrowing = 0;

        this.add(new Phaser.Sprite(game, 0, this.game.groundY, this.rootBitmap));
        this.add(new Phaser.Sprite(game, 0, 0, this.stemBitmap));
        this.drawRoots();
        this.drawStem();
    }

    drawRoots() {
        var ctx = this.rootBitmap.context;
        ctx.lineCap = 'round';
        ctx.lineJoin = 'round';
        ctx.clearRect(0, 0, this.game.width, this.game.height);
        ctx.moveTo(this.game.width / 2, 0);

        var totalRoots = 0;
        var usedRoots = 0;

        var drawChildren = (children) => {
            for (var root of children) {
                ctx.strokeStyle = `#${(0xB78139 + parseInt(Math.random() * 100)).toString(16)}`;
                ctx.lineWidth = 2 + 3 / root.depth;

                totalRoots++;
                if (!root.target) continue;
                else usedRoots++;

                ctx.beginPath();
                ctx.moveTo(root.startX, root.startY);

                var percent = root.progress / root.distance;
                if (percent < 1) {
                    root.progress += 1;
                    var angle = Math.atan2((root.endY - root.startY), (root.endX - root.startX));
                    var toX = root.startX + Math.cos(angle) * root.distance * percent;
                    var toY = root.startY + Math.sin(angle) * root.distance * percent;
                    ctx.lineTo(toX, toY);
                } else {
                    ctx.lineTo(root.endX, root.endY);
                    root.split(); // will only split once.
                }

                ctx.closePath();
                ctx.stroke();
                if (root.children.length) drawChildren(root.children);
            }
        };
        drawChildren(this.roots);
        this.rootBitmap.dirty = true;

        this.game.textLines[1] = `Roots used: ${usedRoots}/${totalRoots}`;
    }

    priceLeaf() {
        this.game.costText.setText(`Cost: \n${15} energy\n${10} water`);
    }

    growLeaf() {
        var energyCost = 15;
        var waterCost = 10;
        if (this.leafPositions.length === 0 || this.game.stats.energy < energyCost || this.game.stats.water < waterCost) {
            this.game.add.audio('error').play();
            return;
        }

        var leafPos = this.leafPositions.shift();
        var leaf = this.game.add.sprite(leafPos[0], leafPos[1], 'leaf');
        leaf.update = function () {
            if (this.baseScale < 0.4) {
                this.baseScale += 0.02;
            }
            this.scale.x = this.baseScale;
            this.scale.y = this.baseScale;
        };
        leaf.anchor.x = 1.0;
        leaf.anchor.y = 0.5;
        leaf.scale.y = 0.4;
        leaf.baseScale = 0;
        leaf.animSeed = Math.random();
        leaf.rotation = leafPos[2];
        leaf.bringToTop();
        this.game.add.audio('build').play();
        this.game.stats.leaves++;
        this.game.stats.waterPerDay--;
        this.game.stats.energyPerDay++;

        this.game.stats.energy -= energyCost;
        this.game.stats.water -= waterCost;
    }

    findStem(rootStem) {
        var depth = 5;
        var lowestStem;
        var findSplit = (stems) => {
            for (var stem of stems) {
                if (stem.depth < depth && stem.children.length === 0) {
                    depth = stem.depth;
                    lowestStem = stem;
                }
                findSplit(stem.children);
            }
        };
        findSplit(rootStem);

        return [lowestStem, depth];
    }

    priceStem() {
        var stem = this.findStem(this.stem);
        if (stem[0]) {
            this.game.costText.setText(`Cost: \n${5 + stem[1]} energy\n${5 + stem[1] * 2} water`);
        }
    }

    growStem() {
        console.log(this.stem);
        var stem = this.findStem(this.stem);

        var energyCost = 5 + stem[1];
        var waterCost = 5 + stem[1] * 2;
        this.stemsGrowing += 3;

        if (stem[0] && this.game.stats.energy >= energyCost && this.game.stats.water >= waterCost) {
            stem[0].split();
            this.game.stats.energy -= energyCost;
            this.game.stats.water -= waterCost;
            this.game.add.audio('build').play();
        } else {
            this.game.add.audio('error').play();
        }
    }

    drawStem() {
        var ctx = this.stemBitmap.context;
        ctx.lineCap = 'round';
        ctx.lineJoin = 'round';
        ctx.clearRect(0, 0, this.game.width, this.game.height);
        ctx.moveTo(this.game.width / 2, 0);

        var drawChildren = (children) => {
            var numSplit = children.length;
            var i = 0;
            for (var stem of children) {
                ctx.strokeStyle = `#08${(160 - Math.random() * 20).toString(16).substr(0, 2)}00`;
                ctx.lineWidth = 2 + 4 / stem.depth;

                ctx.beginPath();
                ctx.moveTo(stem.startX, stem.startY);
                var finishGrowth = false;

                if (stem.progress < 1) {
                    stem.progress += 0.05;

                    if (stem.progress >= 1) finishGrowth = true;
                }

                var endX;
                var angle = Math.PI * (-numSplit / 2) * i;
                if (stem.depth === 1) {
                    endX = stem.startX
                }
                else {
                    endX = stem.startX + Math.cos(angle) * stem.depth * 10 * stem.progress;
                }

                var yAdd = 35 * stem.progress;
                var endY = stem.startY - Math.sin(angle) * 20 * stem.progress;

                if (finishGrowth) {
                    // Twice
                    var leafX = stem.startX + Math.cos(angle) * stem.depth * 10 * 0.5;
                    var leafY = stem.startY - Math.sin(angle) * 20 * 0.5 - 17.5;

                    if (stem.depth == 1) leafX -= 4;
                    this.leafPositions.push([leafX, leafY, Math.PI / 4 + (Math.random() - 0.5) * Math.PI / 2]);
                    this.leafPositions.push([leafX, leafY, Math.PI - Math.PI / 4 + (Math.random() - 0.5) * Math.PI / 2]);
                    this.stemsGrowing--;
                }

                ctx.lineTo(endX, endY - yAdd);
                if (stem.progress < 1) {
                    stem.progress += 0.025;
                } else {
                    stem.endX = endX;
                    stem.endY = endY - yAdd;
                }


                ctx.closePath();
                ctx.stroke();
                if (stem.children.length) drawChildren(stem.children);
                i++;
            }
        };
        drawChildren(this.stem);
        this.stemBitmap.dirty = true;
    }
}

class Stem {
    constructor(game, parent) {
        this.game = game;
        this.parent = parent;
        this.children = [];

        if (parent) {
            this.startX = parent.endX;
            this.startY = parent.endY;
            this.depth = parent.depth + 1;
        } else {
            this.startX = game.width / 2;
            this.startY = game.height - game.height / 3;
            this.depth = 1;
        }

        this.progress = 0;
        this.done = false;
    }

    split() {
        console.log(this.children.length);
        if (this.progress >= 1 && this.children.length == 0) {
            this.children = [new Stem(this.game, this), new Stem(this.game, this), new Stem(this.game, this)];
            this.done = true;
            return true;
        }
    }
}

class Root {
    constructor(game, parent) {
        this.game = game;
        this.parent = parent;
        this.children = [];
        if (parent) {
            this.startX = parent.endX;
            this.startY = parent.endY;
            this.depth = parent.depth + 1;
        } else {
            this.startX = game.width / 2;
            this.startY = 0;
            this.depth = 1;
        }

        this.progress = 0;
        this.distance = 0;
    }

    setTarget(nutrient) {
        if (this.target) return;
        this.target = nutrient;
        this.target.root = this;

        this.endX = nutrient.x;
        this.endY = nutrient.y;

        this.distance = distance(this.startX, this.startY, this.endX, this.endY);
        this.game.plant.rootsGrowing++;
    }

    split() {
        if (this.target.connected) return false;
        if (this.progress >= this.distance && this.children.length == 0) {
            this.target.connected = true;
            this.target.alive = true;
            this.children = [new Root(this.game, this), new Root(this.game, this)];
            this.game.plant.rootsGrowing--;
            this.game.stats.waterPerDay++;
        }
    }
}